// ***** Optimizer comments:								*****
// ***** decoder requires 32-bit colors						*****
// ***** AMD K6/2-350:	app 30fps							*****





#include <stdio.h>
#include <math.h>
#include <time.h>

#include "DukeDecode.h"
#include "video.h"

int decWidth		=360;		// decoded (display/output) width
extern int quality;
#define decHeight	280		// decoded (display/output) height
#define rawWidth	2268	// input (raw) width
#define	rawHeight	293		// input (raw) height
#define rawOffset	5*rawWidth - 102	// offset of first pixel

// decode (static)
#define Pi			3.14159265358979323846
#define divPi		(0x7fffffffL /Pi)

#define FIX64(n)	((int) ((double)64*n + .5))
#define MAKEQW(n)	(((__int64) (n & 0xffff) << 48) | ((__int64) (n & 0xffff) << 32) | ((__int64) (n & 0xffff) << 16) | (__int64) (n & 0xffff))
#define rp			rawWidth		// pitch

int FcHF0 		((int) (283.7514837L * 0x100000000L /rawWidth));
int FcVF0 		((int) (  -.25L * 0x100000000L));

extern bool debug_mode;
extern char info_text_buffer[1024];

int cbPhases[288];

static void InitializeDecoder();
static void SetPixels(int y);
static void GetColorBurstData( unsigned char *src, int y1,unsigned int *cbPhase,
							   unsigned int *cbHF, unsigned int *cbVF, int* cbOdd );

__int64 GetLumaSamples_1x( unsigned char *src, int y_pos,int lladjust, int brightness, int contrast ,int addr,char shft,unsigned char *screen);
__int64 GetLumaSamples_2x( unsigned char *src, int y_pos,int lladjust, int brightness, int contrast ,int addr,char shft,unsigned char *screen);
static void GetChromaSamples_1x( unsigned char src[],unsigned int cbPhase, unsigned int cbHF, unsigned int cbVF,int saturation, int vsaturation,int y,unsigned char *screen );
static void GetChromaSamples_2x( unsigned char src[],unsigned int cbPhase, unsigned int cbHF, unsigned int cbVF,int saturation, int vsaturation,int y,unsigned char *screen );




	__int64 level64;
	__int64 level642;
	__int64 contrast64;
	__int64 contrast642;


unsigned char color_adjust=240;

extern int border_x,border_y,start_y2,resolution,saving;

char shift=13;
extern int pitch;
extern int frame_id,frame_id_prev,waiting;
extern unsigned long *fid;

extern unsigned char* pixels;

int craita=13,braita=12;

__int64 getMicroSec(void);
static void interpolate1(void *src);

static void SetPixels_interlace(unsigned long *pixptr,int y);
static void SetPixels_interlace2(unsigned long *pixptr,int y);
static void CalcPixels_Video(void *pixels,int y);
static void CalcPixels_Videox2(void *pixels,int y);

extern int decode_mode;
extern unsigned int frame;
extern int frames_in_buffer,current_frame,frame_flip;
extern __int64 nowms,startms;


int interpolation=0,saturation_correction=0;
extern long unsigned int raidat;

extern int lladjust2;
int lladjust4=0;
unsigned char lladjust3=0;
int bw_mode=0;
extern int lock_x;
extern bool isWindowed;

extern bool disable_ints;

unsigned char	scanLuma[2*720*288];
static int				cosint[1024];
static bool				isInitialized = false;
int correct=0;

__int64 cosin64[1024], prevScan[384];
static short int filt_cos[] = { 141, 199, -282, -1020, -457, 1078, 1559,-704,
							-704, -1559, -1078, 457, 1020, 282, -199, 141 };
static short int filt_sin[] = { -340, 199, 962, 821, -2182, -1525, 364, 1700,
							1700, 364, -1525, -2182, 821, 962, 199, -340 };

dukedecoder::dukedecoder()
{
	_description = "Duke's MMX decoder";
	if (resolution==1) {
		_outwidth = decWidth;
		_outheight = 280;
	} else {
		_outwidth = 720;
		_outheight = 280;
	}
	_use_vbi = true;
}



// Decode next field of PAL data

static long show_settings_timeout;

bool adjust_luma_level = true;

typedef struct CLR_CFG
{
	int brightness;
	int contrast;
	int saturation;
} CLR_CFG;

struct CLR_CFG color_cfg[2] = { {256,256,256}, {-256,-256,-256} };

int color_cfg_i;

void dukedecoder::processKey(WPARAM key)
{
	bool show_it = false;

	switch(key)
   {
   	case VK_PRIOR:
        	show_it = true;
      	if (color_cfg[color_cfg_i].brightness < 511)
				color_cfg[color_cfg_i].brightness+=1;
         break;
      case VK_NEXT:
        	show_it = true;
      	if (color_cfg[color_cfg_i].brightness > -511)
				color_cfg[color_cfg_i].brightness-=1;
         break;
   	case VK_HOME:
        	show_it = true;
      	if (color_cfg[color_cfg_i].contrast < 511)
				color_cfg[color_cfg_i].contrast+=1;
         break;
      case VK_END:
        	show_it = true;
      	if (color_cfg[color_cfg_i].contrast > -511)
				color_cfg[color_cfg_i].contrast-=1;
         break;
   	case VK_INSERT:
        	show_it = true;
      	if (color_cfg[color_cfg_i].saturation < 511)
				color_cfg[color_cfg_i].saturation+=1;
         break;
      case VK_DELETE:
        	show_it = true;
      	if (color_cfg[color_cfg_i].saturation > -511)
				color_cfg[color_cfg_i].saturation-=1;
         break;

		case 'A':
      	adjust_luma_level = !adjust_luma_level;
         break;
		case 'P':
		interpolation=interpolation^1;
         break;
		case '7':
      	FcHF0=FcHF0+16000;
         break;
		case '8':
      	FcHF0=FcHF0+16000;
         break;
		case '9':
      	lladjust3=lladjust3^1;
         break;
		case '3':
         break;
		case '4':
         break;
		case '0':
      	lladjust2=0;
         break;
		case 'Q':
      	lladjust2+=1;
         break;
		case 'W':
      	lladjust2-=1;
         break;
	
		case VK_F3:
			color_cfg[color_cfg_i].saturation=-color_cfg[color_cfg_i].saturation;
		break;

   }

   if (show_it)
   	show_settings_timeout = time(0) + 5;
}

const char* dukedecoder::getHelp(void) const
{
	return("  PG UP/DOWN: Adjust brightness\n"
          "  HOME/END: Adjust contrast\n"
          "  INSERT/DELETE: Adjust saturation\n"
          "  A: Adjust luma level to color burst\n"
          );
}


void dukedecoder::decodeFrame(unsigned char *src, int offset, bool invert_polarity)
{
	unsigned int cbPhase, cbHF, cbVF;
	int x,y,z, vsaturation, cbOdd;
	unsigned char *src3;
	unsigned char *s0;
	unsigned char *s1;
	int src_add;


	src3=src;


	color_cfg_i = (invert_polarity)?1:0;
	int brightness = color_cfg[color_cfg_i].brightness;
	int contrast = color_cfg[color_cfg_i].contrast;
	int saturation = color_cfg[color_cfg_i].saturation;
	bool lladjust = adjust_luma_level;


	if (!isInitialized) {
		InitializeDecoder();
		}


	if ((src==NULL) && (offset==-1)) goto jump0;




	src += offset;
	GetColorBurstData(src, 0,&cbPhase, &cbHF, &cbVF, &cbOdd);
	src_add = getColorBurstWidth();
	vsaturation	= (cbOdd ? -saturation:saturation);
	cbHF	   *= 10;

	


	// ***** Optimizer comments:									*****
	// ***** optional disable interrupts while doing assembler		*****
	// ***** memory operations                                     	*****
	if ((disable_ints) && (!isWindowed)) _asm cli;


	lockDisplay();
	s1 = (unsigned char*)pixels;
	s0=s1+pitch*288;

	src += rawWidth*2;
	cbPhase	+= cbVF*2;
	s1+=pitch*2;

	for (y=(border_y/2)*2+2;y < decHeight+(start_y2/2)*2; y+=2)
	{
		if (y>=decHeight) break;

		if (quality==0) {
			GetChromaSamples_1x( src+src_add, cbPhase, cbHF, cbVF, saturation, vsaturation,y/2*180,s1);
			GetLumaSamples_1x(src+src_add, y,lladjust, brightness, contrast,y*360,1,s1);
			GetLumaSamples_1x(src+src_add+rawWidth, y+1,lladjust, brightness, contrast,y*360+360,0,s1+pitch);
		} else {
			GetChromaSamples_1x( src+src_add, cbPhase, cbHF, cbVF, saturation, vsaturation,y/2*180,s1);
			GetLumaSamples_2x(src+src_add, y,lladjust, brightness, contrast,y*360,1,s1);
			GetLumaSamples_2x(src+src_add+rawWidth, y+1,lladjust, brightness, contrast,y*360+360,0,s1+pitch);
		}
		_asm emms;

		src += rawWidth*2;
		cbPhase	+= cbVF*2;
		s1+=pitch*2;

		if (s1>s0) break;

		}
   if (debug_mode)
   {
	   drawText(0, 0, info_text_buffer, RGB(255, 255, 255));
   }
	unlockDisplay();

	// ***** enable interrupts after asm memory ops						*****
	if ((disable_ints) && (!isWindowed)) _asm sti;

jump0:



	// ***** Optimizer comments:								*****
	// ***** for DMA sync										*****
	frame_id_prev=*fid;
	frame_id_prev+=*(fid+4);
	frame_id_prev+=*(fid+8);
	frame_id_prev+=*(fid+12);

}


// Initialize decoder

static void InitializeDecoder()
{
	int i;

	for (i=0; i < 1024; i++)
	{
		short c = (int) ((double)0x7fffffff * cos( i*Pi /512.0 )) >> 16;
		short s = (int) ((double)0x7fffffff * sin( i*Pi /512.0 )) >> 16;
		short *v = (short *) &cosin64[i];
		v[0] = s;
		v[1] = c;
		v[2] = c;
		v[3] = ~s;
		cosint[i] = c;

	}

	isInitialized = true;
}

__int64 GetLumaSamples( unsigned char *src, int lladjust, int brightness, int contrast ,int addr,char shft,unsigned char *screen)
{
	__int64 r;
	static __int64 fw384 = MAKEQW( 384 /4 );
	static __int64 fw160 = MAKEQW( 128 /4 );
	static __int64 i4096 = MAKEQW( 4096 /4 );
	static __int64 i8192 = MAKEQW( 8192 /4 );
	static __int64 temp = 0x00ff00ff00ff00ff;


	__asm
	{
		mov			esi , [src]			// load input pointer
		mov			edi , [screen]		// load output pointer
		mov			ebx,offset scanLuma

//		movd		mm6 , [contrast]	// load contrast
//		movd		mm5 , [brightness]	// load brightness level

		mov			ecx , decWidth		// load 4-sample counter
		shr ecx,1
/*
	// luma notchfilter, remove color frequency, downsample
	for (x=0; x < decWidth; x++)
	{
		l  = brightness;
		l += ((src[-3]+src[-2]+src[-1]+src[ 0] + src[1]+src[2]+src[3]+src[4]) * contrast) >> 9;
		scanLuma[x] = (l < 0) ? 0 : ((l > 255) ? 255 : l);
		src += 5;
	}
*/

__pixel0:

		mov eax,[brightness]
		xor edx,edx
		mov dl,[esi][-3]
		add ax,dx
		mov dl,[esi][-2]
		add ax,dx
		mov dl,[esi][-1]
		add ax,dx
		mov dl,[esi][0]
		add ax,dx
		mov dl,[esi][+1]
		add ax,dx
		mov dl,[esi][+2]
		add ax,dx
		mov dl,[esi][+3]
		add ax,dx
		mov dl,[esi][+4]
		add ax,dx
		mul [contrast]
		shr eax,8
		mov [edi][1],al

		add esi,5

		mov eax,[brightness]
		xor edx,edx
		mov dl,[esi][-3]
		add ax,dx
		mov dl,[esi][-2]
		add ax,dx
		mov dl,[esi][-1]
		add ax,dx
		mov dl,[esi][0]
		add ax,dx
		mov dl,[esi][+1]
		add ax,dx
		mov dl,[esi][+2]
		add ax,dx
		mov dl,[esi][+3]
		add ax,dx
		mov dl,[esi][+4]
		add ax,dx
		mul [contrast]
		shr eax,8
		mov [edi][3],al
		
		mov al,[ebx][1]
		mov [edi][2],al
		mov al,[ebx][3]
		mov [edi][0],al

		add esi,5
		add edi,4
		add ebx,4


		dec ecx
		cmp ecx,0
		jnz __pixel0

	}

	return r;
}


__int64 GetLumaSamples_2x( unsigned char *src, int y_pos,int lladjust, int brightness, int contrast ,int addr,char shft,unsigned char *screen)
{
	short r_now,r_prev;
	unsigned char aaa;
	//static __int64 fw384 = MAKEQW( 384 /4 );
	static __int64 fw384 = MAKEQW( 128 /4 );
	static __int64 fw160 = MAKEQW( 128 /4 );
	static __int64 i4096 = MAKEQW( 4096 /4 );
	static __int64 i8192 = MAKEQW( 8192 /4 );
	static __int64 temp = 0x00ff00ff00ff00ff;

	level642=0;

	r_prev=0;

	contrast64=contrast;


	__asm
	{
		movd		mm6 , [contrast]	// load contrast
		punpcklwd	mm6 , mm6			// unpack contrast into WORD's
		punpckldq	mm6 , mm6			// unpack contrast into DWORD's

		pxor		mm7 , mm7			// empty unpack (0) register

		mov			esi , [src]			// load input pointer
		mov			edi , [screen]		// load output pointer
		mov			ebx,offset scanLuma


		movq		mm0 , [esi]			// L0: load samples 0..7

		mov			eax , [lladjust]	// load scan adjustment BOOL

		movd		mm5 , [brightness]	// load brightness level
		punpcklwd	mm5 , mm5			// unpack brightness level into WORD's
		
		or			eax , eax			// luma level scan adjustment ?
		jz			__adjusted			// skip luma adjustment if false


	//////////////////////////////////////////
	//	ADJUST BRIGHTNESS TO COLOR BURST	//
	//										//
	//	{mm1, mm2, mm3, mm4} are free		//
	//	{mm6} is contrast					//
	//	{mm5} is brightness					//
	//////////////////////////////////////////
		xor eax,eax
		xor edx,edx
		mov	dl,[esi][-20]
		add ax,dx
		mov	dl,[esi][-21]
		add ax,dx
		mov	dl,[esi][-22]
		add ax,dx
		mov	dl,[esi][-23]
		add ax,dx
		mov	dl,[esi][-24]
		add ax,dx
		mov	dl,[esi][-25]
		add ax,dx
		mov	dl,[esi][-26]
		add ax,dx
		mov	dl,[esi][-27]
		add ax,dx
		mov	dl,[esi][-28]
		add ax,dx
		shr ax,3
		xor ah,ah

		mov r_now,ax
		add ax,r_prev
		shr ax,1


		mov byte ptr [level642][0],al
		mov byte ptr [level642][1],ah
		mov byte ptr [level642][2],al
		mov byte ptr [level642][3],ah
		mov byte ptr [level642][4],al
		mov byte ptr [level642][5],ah
		mov byte ptr [level642][6],al
		mov byte ptr [level642][7],ah

		jmp __adjusted


		movq		mm1 , [esi][-2268-36]	// load luma level samples 0..7
		; slot

		movq		mm2 , mm1			// copy for samples 4..7
		punpcklbw	mm1 , mm7			// unpack samples 0..3 into WORD's

		movq		mm3 , [esi][0*2268-44]	// load luma level samples 8..15
		punpckhbw	mm2 , mm7			// unpack samples 4..7 into WORD's

		paddw		mm1 , mm2			// add samples --> 0..3
		movq		mm4 , mm3			// copy for samples 12..15

		movq		mm2 , [esi][1*2268-52]	// load luma level samples 16..23
		punpcklbw	mm3 , mm7			// unpack samples 8..11 into WORD's

		paddw		mm1 , mm3			// add samples --> 0..3
		punpckhbw	mm4 , mm7			// unpack samples 12..15 into WORD's

		movq		mm3 , mm2			// copy for samples 20..23
		punpcklbw	mm2 , mm7			// unpack samples 16..19 into WORD's

		punpckhbw	mm3 , mm7			// unpack samples 20..23 into WORD's
		paddw		mm1 , mm4			// add samples --> 0..3

		movq		mm4 , [esi][3*2268-146]	// load luma level samples 24..31
		paddw		mm1 , mm2			// add samples --> 0..3

		movq		mm2 , mm4			// copy for samples 28..31
		punpcklbw	mm4 , mm7			// unpack samples 24..27 into WORD's

		paddw		mm1 , mm3			// add samples --> 0..3
		punpckhbw	mm2 , mm7			// unpack samples 28..31 into WORD's

		movq		mm3 , [fw384]		// load constant '384' (4 x WORD)
		paddw		mm1 , mm4			// add samples --> 0..3

		// 11CLK's (non-pairing instructions)
		paddw		mm1 , mm2			// add samples --> 0..3
		psubw		mm3 , mm1			//  (384 - cbluma32)
		pmaddwd		mm3 , mm6			//  (384 - cbluma32) * contrast
		psrad		mm3 , 11			// ((384 - cbluma32) * contrast) >> 11
		punpckldq	mm1 , mm3			// put low part of mm3 in high part of mm1
		paddd		mm3 , mm1			// add low part to high part
		punpckhdq	mm3 , mm3			// copy luminance adjustment as DWORD's
		packssdw	mm3 , mm3			// pack luminance adjustment into WORD's
		paddw		mm5 , mm3			// adjust brightness to color burst

	//////////////////////////////////////////
	//			FILTER LUMINANCE DATA		//
	//////////////////////////////////////////

__adjusted:
		movq		[contrast64] , mm6	//#store contrast (4 x WORD)

		punpckldq	mm5 , mm5			//#unpack brightness level into DWORD's

		movq		[level64] , mm5		//#store brightness level (4 x WORD)
		movq		mm1 , mm0			// L0: copy for samples 4..7

		mov			ecx , decWidth		// load 4-sample counter
		shr ecx,2


		//sub esi,20+2268

		punpcklbw	mm0 , mm7			// L0: unpack samples 0..3

		movq		mm2 , [esi][2]		// L1: load samples 0..7
		//movq		mm2 , [esi][3]		// L1: load samples 0..7
		punpckhbw	mm1 , mm7			// L0: unpack samples 4..7

		paddw		mm0 , mm1			// L0: add samples --> 0..3
		movq		mm3 , mm2			// L1: copy for samples 4..7

		punpcklbw	mm2 , mm7			// L1: unpack samples 0..3
		movq		mm1 , mm0			// L01: copy of samples L0

		movq		mm4 , [esi][4]		// L2: load samples 0..7
		//movq		mm4 , [esi][6]		// L2: load samples 0..7
		punpckhbw	mm3 , mm7			// L1: unpack samples 4..7

__pixel0:
		paddw		mm2 , mm3			// L1: add samples --> 0..3
		movq		mm5 , mm4			// L2: copy for samples 4..7

		movq		mm6 , [esi][6]		// L3: load samples 0..7
		//movq		mm6 , [esi][9]		// L3: load samples 0..7
		punpckldq	mm0 , mm2			// L01: unpack samples 0..1 of L0, L1

		punpckhdq	mm1 , mm2			// L01: unpack samples 2..3 of L0, L1
		movq		mm3 , mm6			// L3: copy for samples 4..7

		paddw		mm0 , mm1			// L01: add samples --> 0..1
		punpcklbw	mm4 , mm7			// L2: unpack samples 0..3

		pmaddwd		mm0 , [contrast64]	// L01: mac samples --> DWORD

		punpckhbw	mm5 , mm7			// L2: unpack samples 4..7

		paddw		mm4 , mm5			// L2: add samples --> 0..3
		punpcklbw	mm6 , mm7			// L3: unpack samples 0..3

		movq		mm5 , mm4			// L23: copy of samples L2
		punpckhbw	mm3 , mm7			// L3: unpack samples 4..7

		psrad		mm0 , 9				// L01: scale samples to normal level
		paddw		mm6 , mm3			// L3: add samples --> 0..3

		punpckldq	mm4 , mm6			// L23: unpack samples 0..1 of L2, L3
		//add			esi , 20			// increase input pointer
		add			esi , 10			// increase input pointer

		punpckhdq	mm5 , mm6			// L23: unpack samples 2..3 of L2, L3
		movq		mm6 , mm0			// L01: free MMX register mm0

		movq		mm2 , [esi][2]		//L1: load samples 0..7
		//movq		mm2 , [esi][3]		//L1: load samples 0..7

		paddw		mm4 , mm5			// L23: add samples --> 0..1

		pmaddwd		mm4 , [contrast64]	// L23: mac samples --> DWORD

		movq		mm3 , mm2			//L1: copy for samples 4..7

		movq		mm0 , [esi][0]		//L0: load samples 0..7
		punpcklbw	mm2 , mm7			//L1: unpack samples 0..3

		movq		mm1 , mm0			//L0: copy for samples 4..7
		punpcklbw	mm0 , mm7			//L0: unpack samples 0..3

		psrad		mm4 , 9				// L23: scale samples to normal level
		add			edi , 8				// increase output pointer

		packssdw	mm6 , mm4			// L0123: pack samples into WORD's
		; slot

		paddusw		mm6 , [level64]		// L0123: add luminance level
		punpckhbw	mm1 , mm7			//L0: unpack samples 4..7

		movq		mm4 , [esi][4]		//L2: load samples 0..7
		//movq		mm4 , [esi][6]		//L2: load samples 0..7
		paddw		mm0 , mm1			//L0: add samples --> 0..3




		pand		mm6,[temp]
		por			mm6,[ebx]
		psubusb		mm6,[level642]	// L0123: add luminance level
		movq		[edi][-7] , mm6		// L0123: store output samples 0..3


		punpckhbw	mm3 , mm7			//L1: unpack samples 4..7


		add ebx,4

		movq		mm1 , mm0			//L01: copy of samples L0

		dec ecx
		cmp ecx,0
		jnz __pixel0

	}

//	rs[y_pos][frame&255]=r_now;

	return 0;
}


// Decode single scanline of Raw 8bit samples in PAL format to chroma (U,V)
// samples at half resolution.

/*
 Decode single scanline of Raw 8bit samples in PAL format to chroma (U,V)
 samples at half resolution.
*/
static void GetChromaSamples_2x( unsigned char src[],
							  unsigned int cbPhase, unsigned int cbHF, unsigned int cbVF,
							  int saturation, int vsaturation ,int y,unsigned char *screen)
{
	__int64 invertv64, saturation64;
	__int64 *prev = prevScan;

	__asm
	{
		movd		mm0 , [saturation]
		pcmpeqd		mm1 , mm1

		mov			esi , [src]			// load input pointer
		psllq		mm1 , 32

		punpckldq	mm0 , [vsaturation]
		; slot

		movq		[invertv64] , mm1
		pxor		mm0 , mm1

		mov			edx , [cbPhase]		// load color burst phase
		pslld		mm0 , 16

		mov			edi , offset scanLuma		// load chrominance V pointer
		//mov			edi , [screen]

		movq		[saturation64] , mm0
		; slot


	//////////////////////////////////////////
	//	DECODE CHROMINANCE COMPONENT OF PAL	//
	//										//
	//	{esi} is source, {edx} is cbPhase	//
	//	{ebx} is CbU[], {edi} is CbV		//
	//////////////////////////////////////////
		mov			ecx , decWidth	// load 2-pixel counter
		shr ecx,1

		movq		mm0 , [esi][-7]		// C0: load samples 0..7
		pxor		mm7 , mm7			// empty unpack (0) register

		movq		mm2 , [esi][1]		// C0: load samples 8..15
		movq		mm1 , mm0			// C0: copy for samples 4..7

		movq		mm3 , mm2			// C0: copy for samples 12..15
		punpcklbw	mm0 , mm7			// C0: unpack samples 0..3 into WORD's

		movq		mm4 , mm0			// C0: copy of samples 0..3 for "sine" filter
		punpckhbw	mm1 , mm7			// C0: unpack samples 4..7 into WORD's

		pmaddwd		mm0 , [filt_cos]	// C0: "cosine" filter samples 0..3
		movq		mm5 , mm1			// C0: copy of samples 4..7 for "sine" filter

		pmaddwd		mm1 , [filt_cos][8]	// C0: "cosine" filter samples 4..7
		punpcklbw	mm2 , mm7			// C0: unpack samples 8..11

		pmaddwd		mm4 , [filt_sin]	// C0: "sine" filter samples 0..3
		movq		mm6 , mm2			// C0: copy of samples 8..11 for "sine" filter

		xor ebx,ebx

__pixel:
		pmaddwd		mm2 , [filt_cos][16]// C0: "cosine" filter samples 8..11
		punpckhbw	mm3 , mm7			// C0: unpack samples 12..15

		pmaddwd		mm5 , [filt_sin][8]	// C0: "sine" filter samples 4..7
		paddd		mm0 , mm1			// C0: add "cosine" filtered samples 0..7

		pmaddwd		mm6 , [filt_sin][16]// C0: "sine" filter samples 8..11
		movq		mm1 , mm3			// C0: copy of samples 12..15 for "sine" filter

		pmaddwd		mm3 , [filt_cos][24]// C0: "cosine" filter samples 12..15
		paddd		mm0 , mm2			// C0: add "cosine" filtered samples 0..11

		pmaddwd		mm1 , [filt_sin][24]// C0: "sine" filter samples 12..15
		paddd		mm4 , mm5			// C0: add "sine" filtered samples 0..7

		paddd		mm4 , mm6			// C0: add "sine" filtered samples 0..11
		mov			eax , edx			// copy of color burst phase

		paddd		mm0 , mm3			// C0: add "cosine" filtered samples 0..15
		shr			eax , 22			// (cb >> 22) ... index into (co)sine table

		paddd		mm4 , mm1			// C0: add "sine" filtered samples 0..15
		punpckldq	mm3 , mm0			// C0: copy low part "cosine" into high part mm3

		punpckldq	mm1 , mm4			// C0: copy low part "sine" into high part mm1
		paddd		mm3 , mm0			// C0: "cosine" filtered data in high DWORD

		movq		mm0 , [rp][esi][-7]	// C1: load samples 0..7
		paddd		mm1 , mm4			// C0: "sine" filtered data in high DWORD

		movq		mm2 , [rp][esi][1]	// C1: load samples 8..15
		punpckhdq	mm3 , mm1			// C0: low DWORD={cosine}, high DWORD={sine}

		psrad		mm3 , 12			// C0: scale (co)sine filtered data
		movq		mm1 , mm0			// C1: copy for samples 4..7

		packssdw	mm3 , mm3			// C0: pack co(sine) filtered data into WORD's
		movq		mm6 , mm2			// C1: copy for samples 12..15

		pmaddwd		mm3 , [cosin64][8*eax]// C0: low DWORD={u}, high DWORD={v}
		punpcklbw	mm0 , mm7			// C1: unpack samples 0..3 into WORD's

		movq		mm4 , mm0			// C1: copy of samples 0..3 for "sine" filter
		punpckhbw	mm1 , mm7			// C1: unpack samples 4..7 into WORD's

		pmaddwd		mm0 , [filt_cos]	// C1: "cosine" filter samples 0..3
		movq		mm5 , mm1			// C1: copy of samples 4..7 for "sine" filter

		pmaddwd		mm1 , [filt_cos][8]	// C1: "cosine" filter samples 4..7
		punpcklbw	mm2 , mm7			// C1: unpack samples 8..11

		pmaddwd		mm4 , [filt_sin]	// C1: "sine" filter samples 0..3
		punpckhbw	mm6 , mm7			// C1: unpack samples 12..15

		pmaddwd		mm5 , [filt_sin][8]	// C1: "sine" filter samples 4..7
		movq		mm7 , mm2			// C1: copy of samples 8..11 for "sine" filter

		pmaddwd		mm2 , [filt_cos][16]// C1: "cosine" filter samples 8..11
		paddd		mm0 , mm1			// C1: add "cosine" filtered samples 0..7

		pmaddwd		mm7 , [filt_sin][16]// C1: "sine" filter samples 8..11
		movq		mm1 , mm6			// C1: copy of samples 12..15 for "sine" filter

		pmaddwd		mm6 , [filt_cos][24]// C1: "cosine" filter samples 12..15
		paddd		mm0 , mm2			// C1: add "cosine" filtered samples 0..11

		pmaddwd		mm1 , [filt_sin][24]// C1: "sine" filter samples 12..15
		paddd		mm4 , mm5			// C1: add "sine" filtered samples 0..7

		paddd		mm4 , mm7			// C1: add "sine" filtered samples 0..11
		mov			eax , [cbVF]		// load color burst vertical stepsize

		paddd		mm0 , mm6			// C1: add "cosine" filtered samples 0..15
		add			eax , edx			// copy of color burst phase

		paddd		mm4 , mm1			// C1: add "sine" filtered samples 0..15
		shr			eax , 22			// (cb >> 22) ... index into (co)sine table

		pxor		mm3 , [invertv64]	// C0: invert chrominace "V" (feature of PAL)
		punpckldq	mm5 , mm0			// C1: copy low part "cosine" into high part mm5

		punpckldq	mm7 , mm4			// C1: copy low part "sine" into high part mm7
		paddd		mm5 , mm0			// C1: "cosine" filtered data in high DWORD

		paddd		mm7 , mm4			// C1: "sine" filtered data in high DWORD

		movq		mm6 , [cosin64][8*eax]// load (co)sine table entry
		punpckhdq	mm5 , mm7			// C1: low DWORD={cosine}, high DWORD={sine}

		mov			eax , [prev]		// previous scanline chrominance pointer
		psrad		mm5 , 12			// C1: scale (co)sine filtered data

		packssdw	mm5 , mm5			// C1: pack co(sine) filtered data into WORD's

		pmaddwd		mm5 , mm6			// C1: low DWORD={u}, high DWORD={v}
		movq		mm6 , [eax]			// load previous scanline chrominance samples

		add			eax , 8				// increase previous chrominance pointer
		add			esi , 10				// increase input pointer

		pxor		mm7 , mm7			//empty unpack (0) register
		mov			[prev] , eax		// store previous chrominance pointer

		movq		mm0 , [esi][-7]		//C0: load samples 0..7
		paddd		mm5 , mm3			// C01: low DWORD={u}, high DWORD={v}

		movq		mm2 , [esi][1]		//C0: load samples 8..15
		paddd		mm6 , mm5			// average chrominance scanlines (PAL feature)

		pmaddwd		mm6 , [saturation64]// adjust saturation levels
		movq		mm1 , mm0			//C0: copy for samples 4..7

		movq		[eax][-8] , mm5		// store current scanline chrominance samples
		movq		mm3 , mm2			//C0: copy for samples 12..15

		mov			eax , [cbHF]		// load color burst horizontal stepsize
		punpcklbw	mm0 , mm7			//C0: unpack samples 0..3 into WORD's

		psrad		mm6 , 6				// scale chroma output samples
		add			edx , eax			// increment color burst phase (H)

		movq		mm4 , mm0			//C0: copy of samples 0..3 for "sine" filter
		packssdw	mm6 , mm6			// pack chroma output samples into WORD's

		pmaddwd		mm0 , [filt_cos]	//C0: "cosine" filter samples 0..3
		packsswb	mm6 , mm6			// pack chroma output samples into BYTE's

		pmaddwd		mm4 , [filt_sin]	//C0: "sine" filter samples 0..3
		punpckhbw	mm1 , mm7			//C0: unpack samples 4..7 into WORD's

		movd		eax , mm6			// copy of chroma output samples
		movq		mm5 , mm1			//C0: copy of samples 4..7 for "sine" filter

		pmaddwd		mm1 , [filt_cos][8]	//C0: "cosine" filter samples 4..7
		punpcklbw	mm2 , mm7			//C0: unpack samples 8..11

		xor			eax , 8080H			// chroma signed to unsigned conversion

		movq		mm6 , mm2			//C0: copy of samples 8..11 for "sine" filter

		add			edi,8

		inc ebx
		cmp ebx,1
		jb nope

		mov			[edi][-7] , ah
		mov			[edi][-5] , al

		xor ebx,ebx
		mov bl,al	
		add bx,[edi][-7-8]
		shr bx,1
		mov [edi][-7-4],bl

		xor ebx,ebx
		mov bl,ah
		add bx,[edi][-5-8]
		shr bx,1
		mov [edi][-5-4],bl

nope:

		dec ecx
		cmp ecx,0
		jnz			__pixel				// do {...} while (counter > 0);

	}
}

__int64 GetLumaSamples_1x( unsigned char *src, int y_pos,int lladjust, int brightness, int contrast ,int addr,char shft,unsigned char *screen)
{
	__int64 r;
	unsigned short int r_now,r_prev;
	unsigned char aaa;
	//static __int64 fw384 = MAKEQW( 384 /4 );
	static __int64 fw384 = MAKEQW( 128 /4 );
	static __int64 fw160 = MAKEQW( 128 /4 );
	static __int64 i4096 = MAKEQW( 4096 /4 );
	static __int64 i8192 = MAKEQW( 8192 /4 );
	static __int64 temp = 0x00ff00ff00ff00ff;

	level642=0;

	r_prev=0;

	contrast64=contrast;


	__asm
	{
		movd		mm6 , [contrast]	// load contrast
		punpcklwd	mm6 , mm6			// unpack contrast into WORD's
		punpckldq	mm6 , mm6			// unpack contrast into DWORD's

		pxor		mm7 , mm7			// empty unpack (0) register

		mov			esi , [src]			// load input pointer
		mov			edi , [screen]		// load output pointer
		mov			ebx,offset scanLuma


		movq		mm0 , [esi]			// L0: load samples 0..7

		mov			eax , [lladjust]	// load scan adjustment BOOL

		movd		mm5 , [brightness]	// load brightness level
		punpcklwd	mm5 , mm5			// unpack brightness level into WORD's
		
		or			eax , eax			// luma level scan adjustment ?
		jz			__adjusted			// skip luma adjustment if false


	//////////////////////////////////////////
	//	ADJUST BRIGHTNESS TO COLOR BURST	//
	//										//
	//	{mm1, mm2, mm3, mm4} are free		//
	//	{mm6} is contrast					//
	//	{mm5} is brightness					//
	//////////////////////////////////////////
/*
			l  = lumo;
			l +=     3 * (src[ 0] + src[1]);
			l +=    -    (src[-1] + src[2]);
			l +=     2 * (src[-3] + src[4]);
			l +=         (src[-4] + src[5]);
			l +=    -    (src[-5] + src[6]);
*/


		xor eax,eax
		xor edx,edx


		add ax,dx
		mov	dl,[esi][-16]
		add ax,dx
		mov	dl,[esi][-17]
		add ax,dx
		mov	dl,[esi][-18]
		add ax,dx
		mov	dl,[esi][-19]
		add ax,dx
		mov	dl,[esi][-20]
		add ax,dx
		mov	dl,[esi][-21]
		add ax,dx
		mov	dl,[esi][-22]
		add ax,dx
		shr ax,3
		xor ah,ah

		//mov r_now,ax
		//add ax,r_prev
		//shr ax,1

		mov byte ptr [level642][0],al
		mov byte ptr [level642][2],al
		mov byte ptr [level642][4],al
		mov byte ptr [level642][6],al
		
		xor eax,eax
		xor edx,edx
		mov	dl,[ebx][-50]
		add ax,dx
		mov	dl,[ebx][-51]
		add ax,dx
		mov	dl,[ebx][-52]
		add ax,dx
		mov	dl,[ebx][-53]
		add ax,dx
		mov	dl,[ebx][-54]
		add ax,dx
		shr ax,2
		mov al,0
		mov byte ptr [level642][1],al
		mov byte ptr [level642][3],al
		mov byte ptr [level642][5],al
		mov byte ptr [level642][7],al
		



__adjusted:
		movq		[contrast64] , mm6	//#store contrast (4 x WORD)

		punpckldq	mm5 , mm5			//#unpack brightness level into DWORD's

		movq		[level64] , mm5		//#store brightness level (4 x WORD)
		movq		mm1 , mm0			// L0: copy for samples 4..7

		mov			ecx , decWidth		// load 4-sample counter
		shr ecx,2


		sub esi,5+2268

		punpcklbw	mm0 , mm7			// L0: unpack samples 0..3

		movq		mm2 , [esi][5]		// L1: load samples 0..7
		punpckhbw	mm1 , mm7			// L0: unpack samples 4..7

		paddw		mm0 , mm1			// L0: add samples --> 0..3
		movq		mm3 , mm2			// L1: copy for samples 4..7

		punpcklbw	mm2 , mm7			// L1: unpack samples 0..3
		movq		mm1 , mm0			// L01: copy of samples L0

		movq		mm4 , [esi][10]		// L2: load samples 0..7
		punpckhbw	mm3 , mm7			// L1: unpack samples 4..7

__pixel0:
		paddw		mm2 , mm3			// L1: add samples --> 0..3
		movq		mm5 , mm4			// L2: copy for samples 4..7

		movq		mm6 , [esi][15]		// L3: load samples 0..7
		punpckldq	mm0 , mm2			// L01: unpack samples 0..1 of L0, L1

		punpckhdq	mm1 , mm2			// L01: unpack samples 2..3 of L0, L1
		movq		mm3 , mm6			// L3: copy for samples 4..7

		paddw		mm0 , mm1			// L01: add samples --> 0..1
		punpcklbw	mm4 , mm7			// L2: unpack samples 0..3

		pmaddwd		mm0 , [contrast64]	// L01: mac samples --> DWORD

		punpckhbw	mm5 , mm7			// L2: unpack samples 4..7

		paddw		mm4 , mm5			// L2: add samples --> 0..3
		punpcklbw	mm6 , mm7			// L3: unpack samples 0..3

		movq		mm5 , mm4			// L23: copy of samples L2
		punpckhbw	mm3 , mm7			// L3: unpack samples 4..7

		psrad		mm0 , 9				// L01: scale samples to normal level
		paddw		mm6 , mm3			// L3: add samples --> 0..3

		punpckldq	mm4 , mm6			// L23: unpack samples 0..1 of L2, L3

		add			esi , 20			// increase input pointer

		punpckhdq	mm5 , mm6			// L23: unpack samples 2..3 of L2, L3
		movq		mm6 , mm0			// L01: free MMX register mm0

		movq		mm2 , [esi][5]		//L1: load samples 0..7
		paddw		mm4 , mm5			// L23: add samples --> 0..1

		pmaddwd		mm4 , [contrast64]	// L23: mac samples --> DWORD

		movq		mm3 , mm2			//L1: copy for samples 4..7

		movq		mm0 , [esi]			//L0: load samples 0..7
		punpcklbw	mm2 , mm7			//L1: unpack samples 0..3

		movq		mm1 , mm0			//L0: copy for samples 4..7
		punpcklbw	mm0 , mm7			//L0: unpack samples 0..3

		psrad		mm4 , 9				// L23: scale samples to normal level
		add			edi , 8				// increase output pointer

		packssdw	mm6 , mm4			// L0123: pack samples into WORD's
		; slot

		paddusw		mm6 , [level64]		// L0123: add luminance level
		punpckhbw	mm1 , mm7			//L0: unpack samples 4..7

		movq		mm4 , [esi][10]		//L2: load samples 0..7
		paddw		mm0 , mm1			//L0: add samples --> 0..3



		pand		mm6,[temp]
		por			mm6,[ebx]
		psubw		mm6,[level642]		// L0123: luminance level
		movq		[edi][-7] , mm6		// L0123: store output samples 0..3


		punpckhbw	mm3 , mm7			//L1: unpack samples 4..7


		add ebx,8




/*

		cmp ecx,45
		jnz jflsd

		xor eax,eax
		mov al,byte ptr [level642]

		mov [edi][-15],al
		mov [edi][-13],al
		mov [edi][-11],al
		mov [edi][-9],al
		mov [edi][-7],al
		mov [edi][-5],al
		mov [edi][-3],al
		mov [edi][-1],al

jflsd:
*/


		movq		mm1 , mm0			//L01: copy of samples L0


		dec ecx
		cmp ecx,0
		jnz __pixel0

	}

	//rs[y_pos][frame]=r_now;

	return r;
}


// Decode single scanline of Raw 8bit samples in PAL format to chroma (U,V)
// samples at half resolution.

/*
 Decode single scanline of Raw 8bit samples in PAL format to chroma (U,V)
 samples at half resolution.
*/
static void GetChromaSamples_1x( unsigned char src[],
							  unsigned int cbPhase, unsigned int cbHF, unsigned int cbVF,
							  int saturation, int vsaturation ,int y,unsigned char *screen)
{
	__int64 invertv64, saturation64;
	__int64 *prev = prevScan;

	__asm
	{
		movd		mm0 , [saturation]
		pcmpeqd		mm1 , mm1

		mov			esi , [src]			// load input pointer
		psllq		mm1 , 32

		punpckldq	mm0 , [vsaturation]
		; slot

		movq		[invertv64] , mm1
		pxor		mm0 , mm1

		mov			edx , [cbPhase]		// load color burst phase
		pslld		mm0 , 16

		mov			edi , offset scanLuma		// load chrominance V pointer
		//mov			edi , [screen]

		movq		[saturation64] , mm0
		; slot


	//////////////////////////////////////////
	//	DECODE CHROMINANCE COMPONENT OF PAL	//
	//										//
	//	{esi} is source, {edx} is cbPhase	//
	//	{ebx} is CbU[], {edi} is CbV		//
	//////////////////////////////////////////
		mov			ecx , decWidth	// load 2-pixel counter
		shr ecx,1

		//movq		mm0 , [esi][-7]		// C0: load samples 0..7
		movq		mm0 , [esi][-7]		// C0: load samples 0..7
		pxor		mm7 , mm7			// empty unpack (0) register

		//movq		mm2 , [esi][1]		// C0: load samples 8..15
		movq		mm2 , [esi][1]		// C0: load samples 8..15
		movq		mm1 , mm0			// C0: copy for samples 4..7

		movq		mm3 , mm2			// C0: copy for samples 12..15
		punpcklbw	mm0 , mm7			// C0: unpack samples 0..3 into WORD's

		movq		mm4 , mm0			// C0: copy of samples 0..3 for "sine" filter
		punpckhbw	mm1 , mm7			// C0: unpack samples 4..7 into WORD's

		pmaddwd		mm0 , [filt_cos]	// C0: "cosine" filter samples 0..3
		movq		mm5 , mm1			// C0: copy of samples 4..7 for "sine" filter

		pmaddwd		mm1 , [filt_cos][8]	// C0: "cosine" filter samples 4..7
		punpcklbw	mm2 , mm7			// C0: unpack samples 8..11

		pmaddwd		mm4 , [filt_sin]	// C0: "sine" filter samples 0..3
		movq		mm6 , mm2			// C0: copy of samples 8..11 for "sine" filter

		xor ebx,ebx

__pixel:
		pmaddwd		mm2 , [filt_cos][16]// C0: "cosine" filter samples 8..11
		punpckhbw	mm3 , mm7			// C0: unpack samples 12..15

		pmaddwd		mm5 , [filt_sin][8]	// C0: "sine" filter samples 4..7
		paddd		mm0 , mm1			// C0: add "cosine" filtered samples 0..7

		pmaddwd		mm6 , [filt_sin][16]// C0: "sine" filter samples 8..11
		movq		mm1 , mm3			// C0: copy of samples 12..15 for "sine" filter

		pmaddwd		mm3 , [filt_cos][24]// C0: "cosine" filter samples 12..15
		paddd		mm0 , mm2			// C0: add "cosine" filtered samples 0..11

		pmaddwd		mm1 , [filt_sin][24]// C0: "sine" filter samples 12..15
		paddd		mm4 , mm5			// C0: add "sine" filtered samples 0..7

		paddd		mm4 , mm6			// C0: add "sine" filtered samples 0..11
		mov			eax , edx			// copy of color burst phase

		paddd		mm0 , mm3			// C0: add "cosine" filtered samples 0..15
		shr			eax , 22			// (cb >> 22) ... index into (co)sine table

		paddd		mm4 , mm1			// C0: add "sine" filtered samples 0..15
		punpckldq	mm3 , mm0			// C0: copy low part "cosine" into high part mm3

		punpckldq	mm1 , mm4			// C0: copy low part "sine" into high part mm1
		paddd		mm3 , mm0			// C0: "cosine" filtered data in high DWORD

		//movq		mm0 , [rp][esi][-7]	// C1: load samples 0..7
		movq		mm0 , [rp][esi][-7]	// C1: load samples 0..7
		paddd		mm1 , mm4			// C0: "sine" filtered data in high DWORD

		//movq		mm2 , [rp][esi][1]	// C1: load samples 8..15
		movq		mm2 , [rp][esi][1]	// C1: load samples 8..15
		punpckhdq	mm3 , mm1			// C0: low DWORD={cosine}, high DWORD={sine}

		psrad		mm3 , 12			// C0: scale (co)sine filtered data
		movq		mm1 , mm0			// C1: copy for samples 4..7

		packssdw	mm3 , mm3			// C0: pack co(sine) filtered data into WORD's
		movq		mm6 , mm2			// C1: copy for samples 12..15

		pmaddwd		mm3 , [cosin64][8*eax]// C0: low DWORD={u}, high DWORD={v}
		punpcklbw	mm0 , mm7			// C1: unpack samples 0..3 into WORD's

		movq		mm4 , mm0			// C1: copy of samples 0..3 for "sine" filter
		punpckhbw	mm1 , mm7			// C1: unpack samples 4..7 into WORD's

		pmaddwd		mm0 , [filt_cos]	// C1: "cosine" filter samples 0..3
		movq		mm5 , mm1			// C1: copy of samples 4..7 for "sine" filter

		pmaddwd		mm1 , [filt_cos][8]	// C1: "cosine" filter samples 4..7
		punpcklbw	mm2 , mm7			// C1: unpack samples 8..11

		pmaddwd		mm4 , [filt_sin]	// C1: "sine" filter samples 0..3
		punpckhbw	mm6 , mm7			// C1: unpack samples 12..15

		pmaddwd		mm5 , [filt_sin][8]	// C1: "sine" filter samples 4..7
		movq		mm7 , mm2			// C1: copy of samples 8..11 for "sine" filter

		pmaddwd		mm2 , [filt_cos][16]// C1: "cosine" filter samples 8..11
		paddd		mm0 , mm1			// C1: add "cosine" filtered samples 0..7

		pmaddwd		mm7 , [filt_sin][16]// C1: "sine" filter samples 8..11
		movq		mm1 , mm6			// C1: copy of samples 12..15 for "sine" filter

		pmaddwd		mm6 , [filt_cos][24]// C1: "cosine" filter samples 12..15
		paddd		mm0 , mm2			// C1: add "cosine" filtered samples 0..11

		pmaddwd		mm1 , [filt_sin][24]// C1: "sine" filter samples 12..15
		paddd		mm4 , mm5			// C1: add "sine" filtered samples 0..7

		paddd		mm4 , mm7			// C1: add "sine" filtered samples 0..11
		mov			eax , [cbVF]		// load color burst vertical stepsize

		paddd		mm0 , mm6			// C1: add "cosine" filtered samples 0..15
		add			eax , edx			// copy of color burst phase

		paddd		mm4 , mm1			// C1: add "sine" filtered samples 0..15
		shr			eax , 22			// (cb >> 22) ... index into (co)sine table

		pxor		mm3 , [invertv64]	// C0: invert chrominace "V" (feature of PAL)
		punpckldq	mm5 , mm0			// C1: copy low part "cosine" into high part mm5

		punpckldq	mm7 , mm4			// C1: copy low part "sine" into high part mm7
		paddd		mm5 , mm0			// C1: "cosine" filtered data in high DWORD

		paddd		mm7 , mm4			// C1: "sine" filtered data in high DWORD

		movq		mm6 , [cosin64][8*eax]// load (co)sine table entry
		punpckhdq	mm5 , mm7			// C1: low DWORD={cosine}, high DWORD={sine}

		mov			eax , [prev]		// previous scanline chrominance pointer
		psrad		mm5 , 12			// C1: scale (co)sine filtered data

		add			edi,4				// increase chrominance "V" pointer

		packssdw	mm5 , mm5			// C1: pack co(sine) filtered data into WORD's

		pmaddwd		mm5 , mm6			// C1: low DWORD={u}, high DWORD={v}
		movq		mm6 , [eax]			// load previous scanline chrominance samples

		add			eax , 8				// increase previous chrominance pointer
		//add			esi , 10				// increase input pointer
		add			esi , 10				// increase input pointer

		pxor		mm7 , mm7			//empty unpack (0) register
		mov			[prev] , eax		// store previous chrominance pointer

		//movq		mm0 , [esi][-7]		//C0: load samples 0..7
		movq		mm0 , [esi][-7]		//C0: load samples 0..7
		paddd		mm5 , mm3			// C01: low DWORD={u}, high DWORD={v}

		//movq		mm2 , [esi][1]		//C0: load samples 8..15
		movq		mm2 , [esi][1]		//C0: load samples 8..15
		paddd		mm6 , mm5			// average chrominance scanlines (PAL feature)

		pmaddwd		mm6 , [saturation64]// adjust saturation levels
		movq		mm1 , mm0			//C0: copy for samples 4..7

		movq		[eax][-8] , mm5		// store current scanline chrominance samples
		movq		mm3 , mm2			//C0: copy for samples 12..15

		mov			eax , [cbHF]		// load color burst horizontal stepsize
		punpcklbw	mm0 , mm7			//C0: unpack samples 0..3 into WORD's

		psrad		mm6 , 6				// scale chroma output samples
		add			edx , eax			// increment color burst phase (H)

		movq		mm4 , mm0			//C0: copy of samples 0..3 for "sine" filter
		packssdw	mm6 , mm6			// pack chroma output samples into WORD's

		pmaddwd		mm0 , [filt_cos]	//C0: "cosine" filter samples 0..3
		packsswb	mm6 , mm6			// pack chroma output samples into BYTE's

		pmaddwd		mm4 , [filt_sin]	//C0: "sine" filter samples 0..3
		punpckhbw	mm1 , mm7			//C0: unpack samples 4..7 into WORD's

		movd		eax , mm6			// copy of chroma output samples
		movq		mm5 , mm1			//C0: copy of samples 4..7 for "sine" filter

		pmaddwd		mm1 , [filt_cos][8]	//C0: "cosine" filter samples 4..7
		punpcklbw	mm2 , mm7			//C0: unpack samples 8..11

		xor			eax , 8080H			// chroma signed to unsigned conversion

		movq		mm6 , mm2			//C0: copy of samples 8..11 for "sine" filter

		add ah,1
		jnc jfdls
		mov ah,255
jfdls:
		mov			[edi][-7] , ah
		mov			[edi][-5] , al
		mov			[edi][-3] , ah
		mov			[edi][-1] , al

		dec ecx
		cmp ecx,0
		jnz			__pixel				// do {...} while (counter > 0);

	}
}

/// Todo :  optimize this (MMX) ...
// Get color burst information

/*
 Get color burst information
*/
static void GetColorBurstData( unsigned char *src, int y1,unsigned int *cbPhase,
							   unsigned int *cbHF, unsigned int *cbVF, int* cbOdd )
{
	double f_amp;
	unsigned int c0, c1, s0, s1, p0, p1, p2, p3, f_cb;
	int i, v,y, x, f_cos, f_sin;


 	// determine phase offsets of colorburst
	for (y=y1; y < decHeight; y++)
	{
		for (x=0, f_cos=0, f_sin=0, f_cb=0; x < 72; x++)
		{
			i      = *src++;
			f_cos += cosint[ f_cb               >> 22] * i;
			f_sin += cosint[(f_cb - 0x40000000) >> 22] * i;
			f_cb  += FcHF0;
		}
		src  += (rawWidth - 72);

		double div = sqrt( (double) f_cos*f_cos + (double) f_sin*f_sin );
		if (div != 0)
			f_amp = 1 / div;
		else
      		f_amp = 10000;

		c1 = (unsigned int) (acos( f_cos * f_amp ) * divPi);
		s0 = (unsigned int) (asin( f_sin * f_amp ) * divPi) + 0x80000000;
		c0 = ~c1;
		s1 = ~s0 + 0x80000000;
		p0 = abs(c0 - s0);  p1 = abs(c0 - s1);
		p2 = abs(c1 - s0);  p3 = abs(c1 - s1);

		if (p0 < p2 && p0 < p3 || p1 < p2 && p1 < p3)  c1 = c0;
		if (p0 < p1 && p0 < p3 || p2 < p1 && p2 < p3)  s1 = s0;

		cbPhases[y] = (s1 >> 1) + (c1 >> 1);
	}

	// Vc polarity detection  (even/odd frame)
	for (i=y1+2, p0=0, p1=0; i < (decHeight - 2); i += 2)
	{
		p0 += abs( cbPhases[i] - cbPhases[i - 1] ) >> 8;
		p1 += abs( cbPhases[i + 1] - cbPhases[i] ) >> 8;
	}
	*cbOdd = (p0 < p1)?1:0;

	// sample rate correction (color burst)
	for (i=y1+8, v=0; i < (decHeight /2); i++)
	{
		v += (cbPhases[(decHeight /2) + i] - cbPhases[i]) >> 8;
	}
	v			= ((v /(decHeight /2)) << 8) /((decHeight /2) - 8);
	*cbHF		= v /rawWidth +  FcHF0;
	*cbVF		= v           +  FcVF0;
	i			= cbPhases[0] +  162 * (*cbHF) - 0x80000000L;
	i		   -= (*cbOdd ? 0 : -0x40000000);
	*cbPhase	= i - *cbHF - (*cbHF >> 1);
}
