#include <windows.h>
#include <ddraw.h>
#include <stdio.h>
#include "video.h"
#include <fcntl.h>
#include <io.h>
#include <sys/stat.h>


// fullscreen window size
#define FS_X	1024
#define FS_Y	768


#define FIX64(n)	((int) (64*n + .5))
#define MAKEQW(n)	(((__int64) (n & 0xffff) << 48) | ((__int64) (n & 0xffff) << 32) | ((__int64) (n & 0xffff) << 16) | (__int64) (n & 0xffff))

// These are the pixel formats this app supports.  Most display adapters 
// with overlay support will recognize one or more of these formats. 
// The first on our list is the 16-bit RGB formats.  These have the widest 
// support. 
DDPIXELFORMAT g_ddpfOverlayFormats[] =  {
{sizeof(DDPIXELFORMAT), DDPF_RGB, 0, 32,  0x7C00, 0x03e0, 0x001F, 0},      
{sizeof(DDPIXELFORMAT), DDPF_FOURCC,MAKEFOURCC('U','Y','V','Y'),0,0,0,0,0},
{sizeof(DDPIXELFORMAT), DDPF_RGB, 0, 16,  0xF800, 0x07e0, 0x001F, 0},   
{sizeof(DDPIXELFORMAT), DDPF_YUV, 0, 0,  0, 0, 0, 0}   };


#define NUM_OVERLAY_FORMATS (sizeof(g_ddpfOverlayFormats) / sizeof(g_ddpfOverlayFormats[0]))  

extern unsigned long int stripe[280];

extern int border_x,border_y,start_y2;


static LPDIRECTDRAW         g_pDD;  		// DirectDraw object
static LPDIRECTDRAWSURFACE  g_pDDSPrimary;	 // DirectDraw primary surface
static LPDIRECTDRAWSURFACE    g_pDDSBack;		// DirectDraw back surface
static RECT                    g_rcWindow;      // Saves the window size & pos.
static RECT                    g_rcViewport;    // Pos. & size to blt from
static RECT                    g_rcScreen;      // Screen pos. for blt

static bool g_bWindowed;

static DDSURFACEDESC      ddsd;

static int bpp;
static int bytespp;
unsigned char* pixels;
int pitch;
static int start_x, max_x;
static int start_y, max_y;
static HWND	myWindow;	// surface window
static int srcwidth;
static int srcheight;

extern int joku_x;

static bool first_time_clear;

static DWORD r_arr[256];
static DWORD g_arr[256];
static DWORD b_arr[256];

static DWORD r_bitmask;
static DWORD g_bitmask;
static DWORD b_bitmask;

extern int horiz_x;



static int InitDdrawSurfaces(bool windowed);


extern bool always_on_top;
extern bool aspect_ratio;

static int r_shift;
static int g_shift;
static int b_shift;

static int r_cnt;
static int g_cnt;
static int b_cnt;

extern bool use_double_buffering;


static void findShiftValue(DWORD mask, int &shft, int& cnt)
{
	int t;
	for (t = 0; t < 32; t++)
   	if (mask & (1<<t))
      	break;
   shft = t;

   cnt = 0;
   while (mask & (1<<t))
   {
   	t++;
      cnt++;
   }
}

static void init_rgb_array(DWORD dwRBitMask, DWORD dwGBitMask, DWORD dwBBitMask)
{
	r_bitmask = dwRBitMask;
	g_bitmask = dwGBitMask;
	b_bitmask = dwBBitMask;

   findShiftValue(r_bitmask, r_shift, r_cnt);
   findShiftValue(g_bitmask, g_shift, g_cnt);
   findShiftValue(b_bitmask, b_shift, b_cnt);

	for (__int64 i = 0; i < 256; i++)
  	{
  		r_arr[i] = (DWORD)(((i * __int64(dwRBitMask)) / 255) & dwRBitMask);
   	g_arr[i] = (DWORD)(((i * __int64(dwGBitMask)) / 255) & dwGBitMask);
  		b_arr[i] = (DWORD)(((i * __int64(dwBBitMask)) / 255) & dwBBitMask);
   }
}

#define MAKERGB(r,g,b) (r_arr[r]|g_arr[g]|b_arr[b])


static void ReleaseAllObjects(void)
{
    if (g_pDD != NULL)
    {
        g_pDD->SetCooperativeLevel(myWindow, DDSCL_NORMAL);
        if (g_pDDSBack != NULL)
        {
            g_pDDSBack->Release();
            g_pDDSBack = NULL;
        }
        if (g_pDDSPrimary != NULL)
        {
            g_pDDSPrimary->Release();
            g_pDDSPrimary = NULL;
        }
    }
}


int initDdraw(HWND hwnd, char* errorbuf)
{
	HRESULT hRet;
   LPDIRECTDRAW    pDD;            // DD1 interface, used to get DD4 interface
   int r;

   myWindow = hwnd;

	if (DirectDrawCreate(NULL, &pDD, NULL) != DD_OK)
   {
   	strcpy(errorbuf, "Couldn't init DirectDraw");
   	return(-1);
   }

    // Fetch DirectDraw4 interface
   hRet = pDD->QueryInterface(IID_IDirectDraw, (LPVOID *)&g_pDD);
   if (FAILED(hRet))
   {
   	strcpy(errorbuf, "QueryInterface FAILED");
   	return(-1);
   }

   // Initialize all the surfaces we need
   r=InitDdrawSurfaces(true);
   if (r == -1)
   {
   	strcpy(errorbuf, "InitSurfaces FAILED");
   	return(-1);
   }
   if (r == -2)
   {
   	strcpy(errorbuf, "Decoder requires 32bit colors.");
   	return(-1);
   }

   return(0);
}


/*
 Free ddraw resources
*/
void uninitDdraw(void)
{
	ReleaseAllObjects();

	if (g_pDD != NULL)
	{
		g_pDD->Release();
		g_pDD = NULL;
	}
}


//-----------------------------------------------------------------------------
// Name: InitSurfaces()
// Desc: Create all the needed DDraw surfaces and set the coop level
//-----------------------------------------------------------------------------
static int InitDdrawSurfaces(bool windowed)
{
	HRESULT		         hRet;
	DDSCAPS            ddscaps;
	LPDIRECTDRAWCLIPPER pClipper;

	g_bWindowed = windowed;

	if (g_bWindowed)
	{
		GetWindowRect(myWindow, &g_rcWindow);

		// Get normal windowed mode
		hRet = g_pDD->SetCooperativeLevel(myWindow, DDSCL_NORMAL);
		if (hRet != DD_OK)
      		return(-1);
		// Create the primary surface
		ZeroMemory(&ddsd,sizeof(ddsd));
		ddsd.dwSize = sizeof(ddsd);
		ddsd.dwFlags = DDSD_CAPS;
		ddsd.ddsCaps.dwCaps =	DDSCAPS_PRIMARYSURFACE|
								DDSCAPS_LOCALVIDMEM|
								DDSCAPS_VIDEOMEMORY;
		hRet = g_pDD->CreateSurface(&ddsd, &g_pDDSPrimary, NULL);
		if (hRet != DD_OK)
      		return(-1);

		// Create a clipper object since this is for a Windowed render
		hRet = g_pDD->CreateClipper(0, &pClipper, NULL);
		if (hRet != DD_OK)
      		return(-1);

		// Associate the clipper with the window
		pClipper->SetHWnd(0, myWindow);
		g_pDDSPrimary->SetClipper(pClipper);
		pClipper->Release();
		pClipper = NULL;

		// Get the backbuffer. For fullscreen mode, the backbuffer was created
		// along with the primary, but windowed mode still needs to create one.
		ddsd.dwFlags        = DDSD_CAPS|DDSD_HEIGHT|DDSD_WIDTH|DDSD_PIXELFORMAT;
		ddsd.ddpfPixelFormat= g_ddpfOverlayFormats[1];
		ddsd.dwWidth        = FS_X;
		ddsd.dwHeight       = FS_Y;
		ddsd.ddsCaps.dwCaps =	DDSCAPS_OVERLAY | DDSCAPS_VIDEOMEMORY;
		hRet = g_pDD->CreateSurface(&ddsd, &g_pDDSBack, NULL);
		if (hRet != DD_OK)
      		return(-1);
	}
	else
    {
		// Get exclusive mode
		hRet = g_pDD->SetCooperativeLevel(myWindow, DDSCL_EXCLUSIVE|DDSCL_FULLSCREEN);
		if (hRet != DD_OK)
      		return(-1);

		// Set the video mode to ...
		hRet = g_pDD->SetDisplayMode( FS_X, FS_Y, 32);   // , 0, 0);
		if (hRet != DD_OK)
      		return(-1);

		SetRect(&g_rcScreen, 0, 0, FS_X, FS_Y);

		// Create the primary surface with 1 back buffer
		ZeroMemory(&ddsd,sizeof(ddsd));
		ddsd.dwSize = sizeof(ddsd);
		ddsd.dwFlags = DDSD_CAPS |
                     DDSD_BACKBUFFERCOUNT|DDSD_REFRESHRATE;
		ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE |
                 			  DDSCAPS_FLIP |
                			  DDSCAPS_COMPLEX |
							  DDSCAPS_LOCALVIDMEM|DDSCAPS_VIDEOMEMORY
							  ;
		ddsd.dwBackBufferCount = 1;
		ddsd.dwRefreshRate= 25;
		hRet = g_pDD->CreateSurface( &ddsd, &g_pDDSPrimary, NULL);
		if (hRet != DD_OK)
      		return(-1);


		// Get the backbuffer. For fullscreen mode, the backbuffer was created
		// along with the primary, but windowed mode still needs to create one.
		ddsd.dwFlags        = DDSD_CAPS|DDSD_HEIGHT|DDSD_WIDTH|DDSD_PIXELFORMAT;
		ddsd.ddpfPixelFormat= g_ddpfOverlayFormats[1];
		ddsd.dwWidth        = FS_X;
		ddsd.dwHeight       = FS_Y;
		ddsd.ddsCaps.dwCaps =	DDSCAPS_OVERLAY | DDSCAPS_VIDEOMEMORY;
		hRet = g_pDD->CreateSurface(&ddsd, &g_pDDSBack, NULL);
		if (hRet != DD_OK)
      		return(-1);
	}

	DDPIXELFORMAT pf;
	memset(&pf, 0, sizeof(pf));
	pf.dwSize = sizeof(pf);
	g_pDDSPrimary->GetPixelFormat(&pf);

	bpp = pf.dwRGBBitCount;
	bytespp = (bpp + 7) / 8;

	//if (bpp!=32) {
	//	return -2;
	//	}

	init_rgb_array(pf.dwRBitMask, pf.dwGBitMask, pf.dwBBitMask);

	return(0);
}


int DdrawChangeCoopLevel(bool windowed)
{
    ReleaseAllObjects();

    if (windowed)
    {
        SetWindowPos(myWindow,
          (always_on_top)?HWND_TOPMOST:HWND_NOTOPMOST,
          g_rcWindow.left, g_rcWindow.top,
          (g_rcWindow.right - g_rcWindow.left),
          (g_rcWindow.bottom - g_rcWindow.top), SWP_SHOWWINDOW );
		ShowCursor(TRUE);
    } else {
	ShowCursor(FALSE);
	}

    int ret = InitDdrawSurfaces(windowed);

    DdrawWindowChanged();  	

    first_time_clear = true;

    return(ret);
}


void DdrawsetSrcSize(int swidth, int sheight)
{
	if (swidth > 0 && sheight > 0)
	{
	 	srcwidth = swidth;
   		srcheight = sheight;
		first_time_clear = true;
	}

	SetRect(&g_rcViewport, border_x, border_y+start_y2, srcwidth-border_x, srcheight-border_y+start_y2);

	if (g_bWindowed)
	{
   		SetWindowPos(myWindow,
			(always_on_top)?HWND_TOPMOST:HWND_NOTOPMOST,
			g_rcWindow.left, g_rcWindow.top,
   			srcwidth+GetSystemMetrics(SM_CXSIZEFRAME)*2,
			srcheight+GetSystemMetrics(SM_CYSIZEFRAME)*2+GetSystemMetrics(SM_CYMENU)*2,
			SWP_SHOWWINDOW);
	}
	DdrawWindowChanged();   
}


void DdrawCheckAspectRatio(void)
{
	if (g_bWindowed)
	{
		GetWindowRect(myWindow, &g_rcWindow);
		int h = g_rcWindow.bottom - g_rcWindow.top
			- GetSystemMetrics(SM_CYSIZEFRAME)*2-GetSystemMetrics(SM_CYMENU)*2;
		int w = h * srcwidth / srcheight;
	
   		SetWindowPos(myWindow,
			(always_on_top)?HWND_TOPMOST:HWND_NOTOPMOST,
			g_rcWindow.left, g_rcWindow.top,
   			w+GetSystemMetrics(SM_CXSIZEFRAME)*2,
			h+GetSystemMetrics(SM_CYSIZEFRAME)*2+GetSystemMetrics(SM_CYMENU)*2,
			SWP_SHOWWINDOW);

   		DdrawWindowChanged();
	}
}


void DdrawTranslateToSrcRect(int& x, int &y)
{
	int h, w;

	if (g_bWindowed)
	{
		RECT rect;
		GetClientRect(myWindow, &rect);
		h = rect.bottom - rect.top;
		w = rect.right - rect.left;
		if (w == 0) w = 1;
		if (h == 0) h = 1;
	}
	else
	{
		w = FS_X;
		h = FS_Y;
	}

	x = (x * srcwidth) / w;
	y = (y * srcheight) / h;
}


void DdrawWindowChanged(void)
{
	if (g_bWindowed)
	{
		GetWindowRect(myWindow, &g_rcWindow);
		GetClientRect(myWindow, &g_rcScreen);
		ClientToScreen(myWindow, (POINT*)&g_rcScreen.left);
		ClientToScreen(myWindow, (POINT*)&g_rcScreen.right);
	}

  	if (use_double_buffering)
	{
   		start_x = 0;
		start_y = 0;
   		max_x = FS_X;
		max_y = FS_Y;
	}
	else
	{
   		if (g_bWindowed)
		{
	   		start_x = g_rcScreen.left;
			start_y = g_rcScreen.top;
		}
		else
		{
      		start_x = (FS_X - srcwidth) / 2;
			start_y = (FS_Y - srcheight) / 2;
		}

		max_x = GetSystemMetrics(SM_CXFULLSCREEN);
		max_y = GetSystemMetrics(SM_CYFULLSCREEN);
	}
}


void DdrawCopyBackToFront(void)
{
	HRESULT hRet;

	if (!use_double_buffering)
   		return;

	while (1)
	{
		hRet = g_pDDSPrimary->Blt(&g_rcScreen, g_pDDSBack,
			&g_rcViewport, DDBLT_WAIT, NULL);
		if (hRet == DD_OK ) 
			break;
		if (hRet == DDERR_SURFACELOST )
		{
			hRet = g_pDDSPrimary->Restore();
			if (hRet != DD_OK )
         		break;
		}
		if (hRet != DDERR_WASSTILLDRAWING )
			break;
	}
}


/*
 Lock display, get pointer to video memory in "pixels"
*/
void lockDisplay(void)
{
	int ddrval;

	ddsd.dwSize = sizeof(ddsd);
	ddsd.dwFlags = DDSD_PITCH;

	while (true)
	{
		if (use_double_buffering) 
			ddrval = g_pDDSBack->Lock(NULL, &ddsd, DDLOCK_WAIT, NULL );
		else
      		ddrval = g_pDDSPrimary->Lock(NULL, &ddsd, DDLOCK_WAIT, NULL );

		if (ddrval == DD_OK)
			break;
		if (ddrval == DDERR_SURFACELOST)
			g_pDDSBack->Restore();
		else
      		exit(-1);
	}

	pitch = ddsd.lPitch;
	pixels = (unsigned char*)ddsd.lpSurface;
}

/*
 Unlock the display (stop drawing)
*/
void unlockDisplay(void)
{
	if (pixels != NULL)
	{
		pixels = NULL;

		if (use_double_buffering)
 			g_pDDSBack->Unlock(NULL);
		else
 			g_pDDSPrimary->Unlock(NULL);
	}
}


void clearDisplay(void)
{
	if (!use_double_buffering && !first_time_clear)
   	return;
   first_time_clear = false;

	// Use the blter to do a color fill to clear the back buffer
   DDBLTFX     ddbltfx;
   ZeroMemory(&ddbltfx, sizeof(ddbltfx));
   ddbltfx.dwSize = sizeof(ddbltfx);
   ddbltfx.dwFillColor = 0;
  	if (use_double_buffering) 
		g_pDDSBack->Blt(NULL, NULL, NULL, DDBLT_COLORFILL | DDBLT_WAIT, &ddbltfx);
	 else
   		g_pDDSPrimary->Blt(NULL, NULL, NULL, DDBLT_COLORFILL | DDBLT_WAIT, &ddbltfx);
}


void putPixel(int x, int y, unsigned char red, unsigned char green, unsigned char blue)
{
   x += start_x;
   y += start_y;
   if (x < 0 || x > max_x || y < 0 || y > max_y)
   	return;

   lockDisplay();

   unsigned char* surface = pixels + pitch * y + x * bytespp;

   if (bpp == 16)
	   *(unsigned short *)surface = (unsigned short)MAKERGB(red, green, blue);
   else if (bpp == 32)
	   *(unsigned long *)surface = (unsigned long)MAKERGB(red, green, blue);

   unlockDisplay();
}


void drawText(int x, int y, const char* s, COLORREF color)
{
	HDC hdc,hdc2;
	HRESULT ret,ret2;

   ret = g_pDDSPrimary->GetDC(&hdc);
   ret2 = g_pDDSBack->GetDC(&hdc2);

   x += start_x;
   y += start_y;

   	SetBkMode(hdc, OPAQUE);
	SetBkColor(hdc,0x007f0000);
      SetTextColor(hdc, color);
      RECT rect;
      rect.left = x;
      rect.top = y+54;
      rect.right = FS_X-x;
      rect.bottom = FS_Y - y;

	if (ret2 == DD_OK)
   {

		DrawText(hdc2, s, strlen(s), &rect, 0);

	g_pDDSBack->ReleaseDC(hdc);
	}
	if (ret == DD_OK)
   {
   		DrawText(hdc, s, strlen(s), &rect, 0);

	g_pDDSPrimary->ReleaseDC(hdc);
	}
}

